#!/bin/ash
#
# Script to configure packet filtering on a small router.
#
# Copyright (c) 2004,2008,2010-2011 Christian Schneider. All rights reserved.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation, not any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# USE ON YOUR OWN RISK!

VERSION="v0.4.1-alpha"

INSTDIR="${0%/*}"

########################################################################
# variables
########################################################################

echo "Firewall: Setting variables ..."

. "${INSTDIR}/variables.conf"

# just print what would be executed, but do not execute anything
[[ -z "${DEBUG}" ]] && DEBUG="0"

# output of executed commands
[[ -z "${OUTPUT}" ]] && OUTPUT="1"

# string at the beginning of the log prefix
[[ -z "${PREFIX}" ]] &&  PREFIX="FIREWALL"

# IP with telnet and SSH access for exit-on-error case
[[ -z "${ADMIN_IP}" ]] &&  ADMIN_IP="192.168.1.0/24"

# log level
[[ -z "${LEVEL}" ]] && LEVEL="info"

# commands with full path
[[ -z "${INSMOD}" ]] && INSMOD="/sbin/insmod"
[[ -z "${IPTABLES}" ]] && IPTABLES="/usr/sbin/iptables"

########################################################################
# debugging mode
########################################################################

if [[ "${DEBUG}" == "1" ]]; then
  echo "**************************************************************"
  echo "WARNING: FIREWALL RUN IN DEBUGGING MODE (NO COMMAND EXECUTED)!"
  echo "**************************************************************"
fi

########################################################################
# functions
########################################################################

echo "Firewall: Loading functions ..."

# load fallback rules and exit
# arguments: no arguments
exit_error () {
  echo "Firewall: Loading fallback rules and exiting ..." >&2

  if [[ "${DEBUG}" != "1" ]]; then
    # flush all tables
    for I in filter nat mangle; do
      "${IPTABLES}" -t "${I}" -F
      "${IPTABLES}" -t "${I}" -X
    done

    # drop all packages
    for I in INPUT FORWARD OUTPUT; do
      "${IPTABLES}" -P "${I}" DROP
    done

    # accept packets over loopback interface
    "${IPTABLES}" -A INPUT -i lo -j ACCEPT
    "${IPTABLES}" -A OUTPUT -o lo -j ACCEPT

    # allow SSH and TELNET from administrator IP(s)
    for PRT in 22 23; do
      "${IPTABLES}" -A INPUT -s "${ADMIN_IP}" -p tcp --dport "${PRT}" -j ACCEPT
      "${IPTABLES}" -A OUTPUT -d "${ADMIN_IP}" -p tcp --sport "${PRT}" -j ACCEPT
    done
  fi

  exit 1
}

# print debug message
# arguments: 1 = message; 2 = error|warning|ok|skipped or exit status
echo_status () {
  case "${2}" in
    o|ok|0)
      [[ "${OUTPUT}" == "1" ]] && echo "  ${1}"
      ;;
    w|warn|warning)
      echo "  *** WARNING: ${1} ***" >&2
      ;;
    s|skip|skipped)
      echo "  *** SKIPPED: ${1} ***" >&2
      ;;
    f|fail|failed)
      echo "  *** FAILED: ${1} ***" >&2
      ;;
    e|error)
      echo "  *** ERROR: ${1} ***" >&2
      exit_error
      ;;
    [1-9]|[1-9][0-9]|[1-9][0-9][0-9])
      echo "  *** ERROR (${2}): ${1} ***" >&2
      exit_error
      ;;
    *)
      echo "  *** UNKNOWN: ${1} ***" >&2
      ;;
  esac
}

# check, if variable(s) is non-empty and do(es) not contain `#'
# arguments: 1 = variable1; 2 = variable2; ...; last = empty
checkpart () {
  ## do not set CURRLINE/CURRFILE, as they should still be valid from
  ## run of checkline function

  # all but last variable
  for CNT in $(seq 2 "${#}"); do
    if [[ "${1}" == "" || "${1#*#}" != "${1}" ]]; then
      if [[ "${CNT}" != "2" ]]; then
        echo_status "line begining with \`${CURRLINE}' in \`${CURRFILE}' incorrect" error
      fi
      return 1
    fi
    shift
  done

  # last variable
  if [[ "${1}" != "" ]]; then
    echo_status "line begining with \`${CURRLINE}' in \`${CURRFILE}' incorrect" error
    return 1
  fi

  # no errors found
  return 0
}

# check, if variable(s) is non-empty and do(es) not contain `#'
# arguments: 1 = filename; 2 = variable1; 3 = variable2; ...; last = empty
checkline () {
  CURRFILE="${1}"
  shift
  CURRLINE="${@}"

  checkpart "${@}"
  return "${?}"
}

# set value of kernel variables to "0" or "1"
set_kernel_vars () {
  VAL=""
  case "${1}" in
    enable)
      VAL="1"
      ;;
    disable)
      VAL="0"
      ;;
    *)
      echo_status "illegal command \`${1}' to set kernel variables" error
      ;;
  esac
  shift

  RETVAL=0
  for I in "${@}"; do
    if [[ "${DEBUG}" == "1" ]] || echo "${VAL}" > "${I}"; then
      echo_status "echo ${VAL} > ${I}" ok
    else
      echo_status "echo ${VAL} > ${I}" error
      RETVAL="$((RETVAL + 1))"
    fi
  done
  return "${RETVAL}"
}

# zone to interface converter
zone2if () {
  while read ZN IF EMPTY; do
    if checkline "interfaces.conf" "${ZN}" "${IF}" "${EMPTY}"; then
      if [[ "${ZN}" == "${1}" ]]; then
        echo "${IF}"
        return 0
      fi
    fi
  done < "${INSTDIR}/interfaces.conf"
  echo_status "zone \`${1}' not found in \`interfaces.conf'" error
  return 1
}

# interface to zone converter
if2zone () {
  while read ZN IF EMPTY; do
    if checkline "interfaces.conf" "${ZN}" "${IF}" "${EMPTY}"; then
      if [[ "${IF}" == "${1}" ]]; then
        echo "${ZN}"
        return 0
      fi
    fi
  done < "${INSTDIR}/interfaces.conf"
  echo_status "interface \`${1}' not found in \`interfaces.conf'" error
  return 1
}

# contruct parameter
# arguments: 1 = parameter, 2 = value
param () {
  if [[ -z "${2}" || "${2}" == "-" ]]; then
    echo
  else
    VAL="${2#\!}"
    if [[ "${2}" == "${VAL}" ]]; then
      echo "${1} ${VAL}"
    else
      echo "! ${1} ${VAL}"
    fi
  fi
  return 0
}

# contruct parameter
# arguments: 1 = l|r|lx|rx, 2 = value1:value2
lrval () {
  LVAL="${2%:*}"
  if [[ "${LVAL}" == "-" ]]; then
    LVAL=""
  fi
  RVAL="${2#*:}"
  if [[ "${RVAL}" == "-" ]]; then
    RVAL=""
  fi

  case "${1}" in
    l)
      echo "${LVAL}"
      ;;
    r)
      echo "${RVAL}"
      ;;
    lx)
      if [[ "${LVAL}" != "${2}" ]]; then
        echo "${LVAL}"
      else
        echo
      fi
      ;;
    rx)
      if [[ "${RVAL}" != "${2}" ]]; then
        echo "${RVAL}"
      else
        echo
      fi
      ;;
    *)
      echo_status "illegal argument \`${1}' of \`lrval' function" error
      ;;
  esac

  return 0
}

# function to call iptables
ipt () {
  if [[ "${DEBUG}" == "1" ]] || "${IPTABLES}" "${@}"; then
    echo_status "iptables ${*}" ok
    return 0
  else
    echo_status "iptables ${*}" error
    return 1
  fi
}

# function to call insmod
ins () {
  if [[ "${DEBUG}" == "1" ]] || "${INSMOD}" "${@}"; then
    echo_status "insmod ${*}" ok
    return 0
  else
    echo_status "insmod ${*}" error
    return 1
  fi
}

########################################################################
# load iptables kernel modules
########################################################################

echo "Firewall: Loading neccessary modules ..."

while read MOD PARAMS; do
  if checkline "modules.conf" "${MOD}" ""; then
    ins "${MOD}" ${PARAMS}
  fi
done < "${INSTDIR}/modules.conf"

########################################################################
# configure kernel variables
########################################################################

echo "Firewall: Configuring kernel variables ..."

while read CMD VAR EMPTY; do
  if checkline "kernel_vars.conf" "${CMD}" "${VAR}" "${EMPTY}"; then
    set_kernel_vars "${CMD}" ${VAR}
  fi
done < "${INSTDIR}/kernel_vars.conf"

########################################################################
# flushing iptables chains
########################################################################

echo "Firewall: Flushing iptables chains ..."

# flush chains and delete user-defined chains
for I in filter nat mangle; do
  ipt -t "${I}" -F
  ipt -t "${I}" -X
done

########################################################################
# loading chains
########################################################################

echo "Firewall: Loading default chain definitions ..."

# log and drop invalid packets
ipt -N noinvalid
ipt -A noinvalid -j LOG --log-prefix "${PREFIX}:invalid:DROP:" --log-level "${LEVEL}"
ipt -A noinvalid -j DROP

# log and drop potential dangerous icmp types
ipt -N dangicmp
ipt -A dangicmp -j LOG --log-prefix "${PREFIX}:dangicmp:DROP:" --log-level "${LEVEL}"
ipt -A dangicmp -j DROP
ipt -N nodangicmp
for I in 5 9 10; do
  ipt -A nodangicmp -p icmp -m icmp --icmp-type "${I}" -j dangicmp
done

# log and drop TCP packets with illegal flag combinations
# (based on shorewall)
ipt -N tcpflags
ipt -A tcpflags -j LOG --log-prefix "${PREFIX}:tcpflags:DROP:" --log-level "${LEVEL}" --log-ip-options
ipt -A tcpflags -j DROP
ipt -N notcpflags
ipt -A notcpflags -p tcp -m tcp --tcp-flags FIN,SYN,RST,PSH,ACK,URG FIN,PSH,URG -j tcpflags
ipt -A notcpflags -p tcp -m tcp --tcp-flags FIN,SYN,RST,PSH,ACK,URG NONE -j tcpflags
ipt -A notcpflags -p tcp -m tcp --tcp-flags SYN,RST SYN,RST -j tcpflags
ipt -A notcpflags -p tcp -m tcp --tcp-flags FIN,SYN FIN,SYN -j tcpflags
ipt -A notcpflags -p tcp -m tcp --sport 0 --tcp-flags SYN,RST,ACK SYN -j tcpflags

# log and drop
ipt -N logdrop
ipt -A logdrop -j LOG --log-prefix "${PREFIX}:logdrop:DROP:" --log-level "${LEVEL}"
ipt -A logdrop -j DROP

# reject in appropriate way
ipt -N reject
ipt -A reject -p tcp -j REJECT --reject-with tcp-reset
ipt -A reject -p udp -j REJECT --reject-with icmp-port-unreachable
ipt -A reject -p icmp -j REJECT --reject-with icmp-host-unreachable
ipt -A reject -j REJECT --reject-with icmp-host-prohibited

# log and reject
ipt -N logreject
ipt -A logreject -j LOG --log-prefix "${PREFIX}:logreject:REJECT:" --log-level "${LEVEL}"
ipt -A logreject -j reject

# default log and drop
ipt -N defltpolicy
ipt -A defltpolicy -j LOG --log-prefix "${PREFIX}:defltpolicy:DROP:" --log-level "${LEVEL}"
ipt -A defltpolicy -j DROP

########################################################################
# basic rules
########################################################################

echo "Firewall: Setting default rules ..."

# accept packets over loopback interface
ipt -A INPUT -i lo -j ACCEPT
ipt -A OUTPUT -o lo -j ACCEPT

# log and drop invalid packets
for I in INPUT FORWARD OUTPUT; do
  ipt -A "${I}" -m state --state INVALID -j noinvalid
done

# log and drop potential dangerous icmp types
for I in INPUT FORWARD OUTPUT; do
  ipt -A "${I}" -p icmp -j nodangicmp
done

# log and drop TCP packets with illegal flag combinations
for I in INPUT FORWARD; do
  ipt -A "${I}" -p tcp -j notcpflags
done

# accept established and related by default
for I in INPUT FORWARD OUTPUT; do
  ipt -A "${I}" -m state --state RELATED,ESTABLISHED -j ACCEPT
done

# set policies
for I in INPUT FORWARD OUTPUT; do
  ipt -P "${I}" DROP
done

########################################################################
# nat zone chains
########################################################################

echo "Firewall: Creating nat zone chains ..."

# input and output chains
while read ZN IF EMPTY; do
  if checkline "interfaces.conf" "${ZN}" "${IF}" "${EMPTY}"; then
    ipt -t nat -N "${ZN}_pre"
    ipt -t nat -A PREROUTING -i "${IF}" -j "${ZN}_pre"
    ipt -t nat -N "${ZN}_post"
    ipt -t nat -A POSTROUTING -o "${IF}" -j "${ZN}_post"
  fi
done < "${INSTDIR}/interfaces.conf"

########################################################################
# user-defined nat rules
########################################################################

echo "Firewall: Setting user-defined nat rules ..."

while read TGT SRC DST PRT TO DPT SPT EMPTY; do
  if checkline "nat.conf" "${TGT}" "${SRC}" "${DST}" "${EMPTY}"; then
    SZN="$(lrval l "${SRC}")"
    DZN="$(lrval l "${DST}")"

    CHN=""
    PARAM=""
    case "${TGT}" in
      DNAT)
        checkpart "${SZN}" "${DZN}"
        CHN="${SZN}_pre"
        PARAM="--to-destination"
        ;;
      REDIRECT)
        checkpart "${SZN}" "${DZN}"
        CHN="${SZN}_pre"
        PARAM="--to-ports"
        ;;
      SNAT)
        checkpart "${DZN}" "${SZN}"
        CHN="${DZN}_post"
        PARAM="--to-source"
        ;;
      MASQUERADE)
        checkpart "${DZN}" "${SZN}"
        CHN="${DZN}_post"
        PARAM="--to-ports"
        ;;
      SAME)
        if [[ ! -z "${DZN}" ]]; then
          checkpart "${DZN}" "${SZN}"
          CHN="${DZN}_post"
        else
          checkpart "${SZN}" "${DZN}"
          CHN="${SZN}_pre"
        fi
        PARAM="--to"
        ;;
      *)
        echo_status "unsupported target \`${TGT}' in NAT rules" error
        ;;
    esac

    SIP="$(param "-s" "$(lrval rx "${SRC}")")"
    DIP="$(param "-d" "$(lrval rx "${DST}")")"
    PROTO="$(param "-p" "${PRT}")"
    DPORT="$(param "--dport" "${DPT}")"
    SPORT="$(param "--sport" "${SPT}")"
    TOPARAM="$(param "${PARAM}" "${TO}")"

    ipt -t nat -A "${CHN}" ${SIP} ${DIP} ${PROTO} ${SPORT} ${DPORT} \
      -j "${TGT}" ${TOPARAM}
  fi
done < "${INSTDIR}/nat.conf"

########################################################################
# tcpmss mangling
########################################################################

echo "Firewall: Mangling of MSS value of TCP SYN packets ..."

# input and output chains
while read ZN EMPTY; do
  if checkline "tcpmss.conf" "${ZN}" "${EMPTY}"; then
    DZN="$(zone2if "${ZN}")"
    ipt -t mangle -A FORWARD -o "${DZN}" -p tcp -m tcp --tcp-flags SYN,RST SYN \
      -j TCPMSS --clamp-mss-to-pmtu
  fi
done < "${INSTDIR}/tcpmss.conf"

########################################################################
# filter zone chains
########################################################################

echo "Firewall: Creating filter zone chains ..."

# input and output chains
while read ZN IF EMPTY; do
  if checkline "interfaces.conf" "${ZN}" "${IF}" "${EMPTY}"; then
    ipt -N "fw2${ZN}"
    ipt -A OUTPUT -o "${IF}" -j "fw2${ZN}"
    ipt -N "${ZN}2fw"
    ipt -A INPUT -i "${IF}" -j "${ZN}2fw"
  fi
done < "${INSTDIR}/interfaces.conf"

# forwarding chains
while read SZN SIF EMPTY1; do
  if checkline "interfaces.conf" "${SZN}" "${SIF}" "${EMPTY1}"; then
    while read DZN DIF EMPTY2; do
      if checkline "interfaces.conf" "${DZN}" "${DIF}" "${EMPTY2}"; then
        if [[ "${SZN}" != "${DZN}" ]]; then
          ipt -N "${SZN}2${DZN}"
          ipt -A FORWARD -i "${SIF}" -o "${DIF}" -j "${SZN}2${DZN}"
        fi
      fi
    done < "${INSTDIR}/interfaces.conf"
  fi
done < "${INSTDIR}/interfaces.conf"

########################################################################
# user-defined filter rules
########################################################################

echo "Firewall: Setting user-defined filter rules ..."

while read TGT SRC DST PRT DPT SPT EMPTY; do
  if checkline "rules.conf" "${TGT}" "${SRC}" "${DST}" "${EMPTY}"; then
    SZN="$(lrval l "${SRC}")"
    DZN="$(lrval l "${DST}")"
    SIP="$(param "-s" "$(lrval rx "${SRC}")")"
    DIP="$(param "-d" "$(lrval rx "${DST}")")"
    PROTO="$(param "-p" "${PRT}")"
    DPORT="$(param "--dport" "${DPT}")"
    SPORT="$(param "--sport" "${SPT}")"

    ipt -A "${SZN}2${DZN}" ${SIP} ${DIP} ${PROTO} ${SPORT} ${DPORT} -j "${TGT}"
  fi
done < "${INSTDIR}/rules.conf"

########################################################################
# user-defined policies
########################################################################

echo "Firewall: Setting user-defined policies ..."

# input and output chains
while read SZN DZN TGT EMPTY; do
  if checkline "policies.conf" "${SZN}" "${DZN}" "${TGT}" "${EMPTY}"; then
    ipt -A "${SZN}2${DZN}" -j "${TGT}"
  fi
done < "${INSTDIR}/policies.conf"

########################################################################
# policies
########################################################################

echo "Firewall: Setting default policies ..."

# default policies
for I in INPUT FORWARD OUTPUT; do
  ipt -A "${I}" -j defltpolicy
done

########################################################################
# raw commands
########################################################################

echo "Firewall: Executing raw commands ..."

. "${INSTDIR}/raw_post.conf"

########################################################################
# debugging mode
########################################################################

if [[ "${DEBUG}" == "1" ]]; then
  echo "**************************************************************"
  echo "WARNING: FIREWALL RUN IN DEBUGGING MODE (NO COMMAND EXECUTED)!"
  echo "**************************************************************"
fi
